#!/usr/bin/env python
import pytesseract
import cv2 as cv
import os
import sys
from math import ceil
import numpy as np
# https://docs.opencv.org/3.2.0/d1/dee/tutorial_moprh_lines_detection.html

def wipeline(a,c,minlength,fg,bg):
    b = (a == fg)
    index = 0
    while True:
        if len(b) == 0:
            break
        first = b.argmax()
        index += first
        b = b[first:]
        length = b.argmin()
        if length == 0:
            if len(a) - index >= minlength:
                c[index:] = bg
            break
        if length >= minlength:
            c[index:index+length] = bg 
        b = b[length:]
        index += length
        
def grabimage():

    global NOISY
    desktop = cv.imread("desktop.png")
    #print("shape desktop:",desktop.shape)
    gray = cv.cvtColor(desktop,cv.COLOR_BGR2GRAY)
    if NOISY:
        cv.imshow("gray",gray)
        cv.imwrite("gray.png",gray)
    (dummy,bw1) = cv.threshold(gray,160,255,cv.THRESH_BINARY)
    bw=bw1
    #print("shape bw:",bw.shape)
    if NOISY:
        cv.imshow("bw1",bw1)
        cv.imwrite("bw1.png",bw1)

    f = open("puzzle_coordinates.in")
    a = f.readline().split()
    x0 = int(a[0])
    y0 = int(a[1])
    a = f.readline().split()
    x1 = int(a[0])
    y1 = int(a[1])
    f.close()

    bwp = bw[y0:y1,x0:x1]
    if NOISY:
        cv.imshow("bwp",bwp)
        cv.imwrite("bwp.png",bwp)

    # maybe the image is color inverted:

    numblack = (bwp == 0).sum()
    numwhite = (bwp == 255).sum()

    if numblack > numwhite:
        print("Inverting image...")
        gray = 255 - gray
        (dummy,bw) = cv.threshold(gray,160,255,cv.THRESH_BINARY)

    if NOISY:
        cv.imshow("bw",bw)
        cv.imwrite("bw.png",bw)
        cv.imshow("gray1",gray)
        cv.imwrite("gray1.png",gray)

    dx = x1-x0
    dy = y1-y0

    # celldimensions:
    cellxmin = int(dx/9)
    cellxmax = ceil(dx/7)
    cellymin = int(dy/9)
    cellymax = ceil(dy/7)

    t = 4
    p = 0.95
    color = 0

# for upper-left, find vertical line at the left, with enough length

    for j in range(x0, x0 - cellxmax, -1):
        xl = -1
        l = (bw[y0:y0+t*cellymin,j] == color).sum()
        if l > p*t*cellymin:
            xl = j
            break
    #print("left vertical line at",xl,y0,"length:",l)

# for upper-left, find horizontal line at the top, with enough length

    for i in range(y0, y0 - cellymax, -1):
        yt = -1
        l = (bw[i,x0:x0+t*cellxmin] == color).sum()
        if l > p*t*cellxmin:
            yt = i
            break
    #print("top horizontal line at",x0,yt,"length:",l)

# for lower-right, find vertical line at the right, with enough length

    for j in range(x1, min(bw.shape[1],x1 + cellxmax), 1):
        xr = -1
        l = (bw[y1:y1-t*cellymin:-1,j] == color).sum()
        if l > p*t*cellymin:
            xr = j
            break
    print("right vertical line at",xr,y1,"length:",l)


# for lower-right, find horizontal line at the bottom, with enough length

    for i in range(y1, min(bw.shape[0],y1 + cellymax), 1):
        yb = -1
        l = (bw[i,x1:x1-t*cellxmin:-1] == color).sum()
        if l > p*t*cellxmin:
            yb = i
            break
    print("lower bottom line at",x1,yb,"length:",l)

    print("coordinates of sudoku:",xl,yt,xr,yb)

    sudoku = bw[yt:yb+1,xl:xr+1]
    if NOISY:
        cv.imshow("sudoku",sudoku)
        cv.imwrite("sudoku.png",sudoku)

    w = sudoku.shape[1]
    h = sudoku.shape[0]
    #print ("w,h:",w,h)
    lw = w/12
    lh = h/12
    im = sudoku.copy()
    for i in range(h):
        wipeline(sudoku[i,:],im[i,:],lw,0,255)
        if True:
            r = w/160
            r = int(np.floor(r/2-r))
            for k in range(r,-r+1):
                if i+k >= 0 and i+k < h:
                    wipeline(sudoku[i,:],im[i+k,:],lw,0,255)
    if NOISY:
        cv.imshow("one",im)
        cv.imwrite("one.png",im)

    for j in range(w):
        wipeline(sudoku[:,j],im[:,j],lh,0,255)
        if True:
            r = h/160
            r = int(np.floor(r/2-r))
            for k in range(r,-r+1):
                if j+k >= 0 and j+k < w:
                    wipeline(sudoku[:,j],im[:,j+k],lh,0,255)
    if NOISY:
        cv.imshow("two",im)
        cv.imwrite("two.png",im)

    return (xl,yt,xr,yb),im

def grabnumbers():

    f = open("number_coordinates.in")
    a = f.readline().split()
    p = [int(a[0]),int(a[1])]
    a = f.readline().split()
    p += [int(a[0]),int(a[1])]
    f.close()
    return p


def getsudoku(image):
    dx = image.shape[0]/9
    dy = image.shape[1]/9
    #print("dx,dy:",dx,dy,image.shape[0],image.shape[1])

    sum_std_dev = 0
    std_devs = []
    sum_cols = []

    # https://github.com/Shreeshrii/tessdata_shreetest/blob/master/digits.traineddata

    try:
        with open("digits.traineddata", "rb") as f:
            byte = f.read(1000000)[999999]
        #custom_config="--psm 10 -c tessedit_char_whitelist=123456789 --oem 0 --tessdata-dir ."
        custom_config="--psm 10 -c tessedit_char_whitelist=123456789 --tessdata-dir . -l digits"
    except:
        custom_config="--psm 10 -c tessedit_char_whitelist=123456789"

    print("custom_config:",custom_config)
    board = np.zeros((9,9),np.int32)
    for i in range(9):
        for j in range(9):
            x1 = i*dx
            y1 = j*dy
            x2 = (i+1)*dx
            y2 = (j+1)*dy
            cell =  image[ceil(x1):ceil(x2),ceil(y1):ceil(y2)]
            if NOISY:
                cv.imwrite("cell"+str(i)+str(j)+".pbm",cell)
            x = pytesseract.image_to_string(cell,config=custom_config)
            try:
                m = int(x)
            except:
                m = 0
            board[i,j] = m
            #print("cell:",i,j,m,ceil(x1),ceil(x2),ceil(y1),ceil(y2))
    return board

def printb(b):
    for i in range(9):
        for j in range(9):
            x = b[i,j]
            if x == 0:
                print(' .',end="")
            else:
                print(' '+str(b[i,j]),end="")
            if j % 3 == 2:
                print('  ',end="")
        print()

def printbf(b,fname):
    f = open(fname,"w")
    for i in range(9):
        for j in range(9):
            print(b[i,j],file=f)
    f.close()

def output_coordinates(corners):
    #print("corners:",corners)
    f = open("coordinates","w")
    dx = (corners[2] -corners[0])/9
    dy = (corners[3] -corners[1])/9
    for j in range(9):
        for i in range(9):
            print(int(corners[0]+i*dx+0.5*dx),\
                  int(corners[1]+j*dy+0.5*dy),file=f)
    f.close()

def output_numbers(corners,numbers):
    #print("corners, numbers:",corners,numbers)
    f = open("numbers","w")
    if (numbers[0] < corners[0] or numbers[0] > corners[2] or \
        numbers[1] < corners[1] or numbers[1] > corners[3]) and \
       (numbers[2] < corners[0] or numbers[2] > corners[2] or \
        numbers[3] < corners[1] or numbers[3] > corners[3]) :
        # the numbers are outside sudoku
        # there are several possibities for the arrangemnt of the numbers:
        #
        # single_row or column:
        # 1 2 3 4 5 6 7 8 9
        #
        # matrix:
        # 1 2 3
        # 4 5 6
        # 7 8 9
        #
        # double_row
        # 1 2 3 4 5
        # 6 7 8 9

        same = 20
        if abs(numbers[1] - numbers[3]) < same or abs(numbers[0] - numbers[2]) < same:
            print("numbers in one row or column")
            dn0 = (numbers[2] - numbers[0])/8
            dn1 = (numbers[3] - numbers[1])/8
            for i in range(9):
                print(round(numbers[0] + i*dn0),round(numbers[1]+i*dn1),file=f)
        elif abs(abs(numbers[0]-numbers[2]) - abs(numbers[1]-numbers[3])) < same:
            print("numbers in 3x3 matrix")
            dn0 = (numbers[2] - numbers[0])/2
            dn1 = (numbers[3] - numbers[1])/2
            for j in range(3):
                for i in range(3):
                    x = numbers[0] + i*dn0
                    y = numbers[1] + j*dn1
                    print(round(x),round(y),file=f)
        else:
            print("assuming numbers in two rows")
            dn0 = (numbers[2] - numbers[0])/3
            dn1 = (numbers[3] - numbers[1])
            for i in range(5):
                print(round(numbers[0] + i*dn0),round(numbers[1]),file=f)
            for i in range(4):
                print(round(numbers[0] + i*dn0),round(numbers[3]),file=f)
    else:
        print("no numbers")
    f.close()

def main():
    global NOISY
    numbers_only = False
    NOISY = False
    arg=""
    if len(sys.argv) > 1:
        arg=sys.argv[1]
        NOISY = True
    if arg == "get":
        os.system("./get_puzzle_coordinates")
        NOISY = True
    elif arg == 'numbers':
        numbers_only = True
        NOISY = False

    corners,image=grabimage()

    if numbers_only:
        os.system("./get_number_coordinates")
        numbers=grabnumbers()
        output_numbers(corners,numbers)
        return

    if NOISY:
        cv.imshow("image",image)
        cv.imwrite("image.png",image)
        cv.waitKey(1)
    board = getsudoku(image)
    #print(board)

    output_coordinates(corners)

    printb(board)
    printbf(board,"sudoku-puzzle")

    sys.stdout.flush()
    if NOISY:
        cv.waitKey(10)
    sys.stdin.read(1)

if __name__ == "__main__":
    try:
        main()
    except:
        import traceback
        print("oeps")
        print(traceback.format_exc())
        sys.stdout.flush()
        if NOISY:
            cv.waitKey(10)
            sys.stdin.read(1)


